@@ -1,4 +1,14 @@
 /*
+ * SPDX-License-Identifier: GPL-2.0-or-later
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
+ *
+ * Copyright (C) 2007 Bernhard Walle <bwalle@suse.de>
+ * Copyright (C) 2007-2023 Karel Zak <kzak@redhat.com>
+ *
  * rtcwake -- enter a system sleep state until specified wakeup time.
  *
  * This uses cross-platform Linux interfaces to enter a system sleep state,
@@ -18,17 +28,19 @@
  * RTC uses a local timezone instead (maybe you dual-boot MS-Windows).
  * That flag should not be needed on systems with adjtime support.
  */
-
 #include <errno.h>
 #include <fcntl.h>
 #include <getopt.h>
 #include <linux/rtc.h>
+#include <poll.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/ioctl.h>
+#include <sys/stat.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <termios.h>
 #include <time.h>
 #include <unistd.h>
 
@@ -63,7 +75,7 @@ enum rtc_modes {	/* manual page --mode option explains these. */
 
 };
 
-static const char *rtcwake_mode_string[] = {
+static const char *const rtcwake_mode_string[] = {
 	[OFF_MODE] = "off",
 	[NO_MODE] = "no",
 	[ON_MODE] = "on",
@@ -84,12 +96,13 @@ struct rtcwake_control {
 	enum clock_modes clock_mode;	/* hwclock timezone */
 	time_t sys_time;		/* system time */
 	time_t rtc_time;		/* hardware time */
-	unsigned int verbose:1,		/* verbose messaging */
-		     dryrun:1;		/* do not set alarm, suspend system, etc */
+	bool	verbose,		/* verbose messaging */
+		dryrun;			/* do not set alarm, suspend system, etc */
 };
 
-static void __attribute__((__noreturn__)) usage(FILE *out)
+static void __attribute__((__noreturn__)) usage(void)
 {
+	FILE *out = stdout;
 	fputs(USAGE_HEADER, out);
 	fprintf(out,
 	      _(" %s [options]\n"), program_invocation_short_name);
@@ -113,13 +126,10 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	fputs(_(" -u, --utc                RTC uses UTC\n"), out);
 	fputs(_(" -v, --verbose            verbose messages\n"), out);
 
-	printf(USAGE_SEPARATOR);
-	printf(USAGE_HELP);
-	printf(USAGE_VERSION);
-
-	printf(USAGE_MAN_TAIL("rtcwake(8)"));
-
-	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
+	fputs(USAGE_SEPARATOR, out);
+	fprintf(out, USAGE_HELP_OPTIONS(26));
+	fprintf(out, USAGE_MAN_TAIL("rtcwake(8)"));
+	exit(EXIT_SUCCESS);
 }
 
 static int is_wakeup_enabled(const char *devname)
@@ -194,21 +204,23 @@ static int get_basetimes(struct rtcwake_control *ctl, int fd)
 		/* Unless the system uses UTC, either delta or tzone
 		 * reflects a seconds offset from UTC.  The value can
 		 * help sort out problems like bugs in your C library. */
-		printf("\tdelta   = %ld\n", ctl->sys_time - ctl->rtc_time);
+		char s[64];
+		printf("\tdelta   = %"PRId64"\n", (int64_t) ctl->sys_time - ctl->rtc_time);
 		printf("\ttzone   = %ld\n", timezone);
 		printf("\ttzname  = %s\n", tzname[daylight]);
+		gmtime_r(&ctl->sys_time, &tm);
+		printf("\tsystime = %"PRId64", (UTC) %s",
+				(int64_t) ctl->sys_time, asctime_r(&tm, s));
 		gmtime_r(&ctl->rtc_time, &tm);
-		printf("\tsystime = %ld, (UTC) %s",
-				(long) ctl->sys_time, asctime(gmtime(&ctl->sys_time)));
-		printf("\trtctime = %ld, (UTC) %s",
-				(long) ctl->rtc_time, asctime(&tm));
+		printf("\trtctime = %"PRId64", (UTC) %s",
+				(int64_t) ctl->rtc_time, asctime_r(&tm, s));
 	}
 	return 0;
 }
 
 static int setup_alarm(struct rtcwake_control *ctl, int fd, time_t *wakeup)
 {
-	struct tm		*tm;
+	struct tm		tm;
 	struct rtc_wkalrm	wake = { 0 };
 
 	/* The wakeup time is in POSIX time (more or less UTC).  Ideally
@@ -220,13 +232,13 @@ static int setup_alarm(struct rtcwake_control *ctl, int fd, time_t *wakeup)
 	 *
 	 * Else clock_mode == CM_LOCAL so the time given to the RTC will
 	 * instead use the local time zone. */
-	tm = localtime(wakeup);
-	wake.time.tm_sec = tm->tm_sec;
-	wake.time.tm_min = tm->tm_min;
-	wake.time.tm_hour = tm->tm_hour;
-	wake.time.tm_mday = tm->tm_mday;
-	wake.time.tm_mon = tm->tm_mon;
-	wake.time.tm_year = tm->tm_year;
+	localtime_r(wakeup, &tm);
+	wake.time.tm_sec = tm.tm_sec;
+	wake.time.tm_min = tm.tm_min;
+	wake.time.tm_hour = tm.tm_hour;
+	wake.time.tm_mday = tm.tm_mday;
+	wake.time.tm_mon = tm.tm_mon;
+	wake.time.tm_year = tm.tm_year;
 	/* wday, yday, and isdst fields are unused */
 	wake.time.tm_wday = -1;
 	wake.time.tm_yday = -1;
@@ -246,12 +258,15 @@ static char **get_sys_power_states(struct rtcwake_control *ctl)
 
 	if (!ctl->possible_modes) {
 		char buf[256] = { 0 };
+		ssize_t ss;
 
 		fd = open(SYS_POWER_STATE_PATH, O_RDONLY);
 		if (fd < 0)
 			goto nothing;
-		if (read(fd, &buf, sizeof(buf) - 1) <= 0)
+		ss = read(fd, &buf, sizeof(buf) - 1);
+		if (ss <= 0)
 			goto nothing;
+		buf[ss] = '\0';
 		ctl->possible_modes = strv_split(buf, " \n");
 		close(fd);
 	}
@@ -262,6 +277,22 @@ nothing:
 	return NULL;
 }
 
+static void wait_stdin(struct rtcwake_control *ctl)
+{
+	struct pollfd fd[] = {
+		{.fd = STDIN_FILENO, .events = POLLIN}
+	};
+	int tries = 0;
+
+	while (tries < 8 && poll(fd, 1, 10) == 1) {
+		if (ctl->verbose)
+			warnx(_("discarding stdin"));
+		xusleep(250000);
+		tcflush(STDIN_FILENO, TCIFLUSH);
+		tries++;
+	}
+}
+
 static void suspend_system(struct rtcwake_control *ctl)
 {
 	FILE	*f = fopen(SYS_POWER_STATE_PATH, "w");
@@ -272,6 +303,8 @@ static void suspend_system(struct rtcwake_control *ctl)
 	}
 
 	if (!ctl->dryrun) {
+		if (isatty(STDIN_FILENO))
+			wait_stdin(ctl);
 		fprintf(f, "%s\n", ctl->mode_str);
 		fflush(f);
 	}
@@ -315,6 +348,7 @@ static int print_alarm(struct rtcwake_control *ctl, int fd)
 	struct rtc_wkalrm wake;
 	struct tm tm = { 0 };
 	time_t alarm;
+	char s[CTIME_BUFSIZ];
 
 	if (ioctl(fd, RTC_WKALM_RD, &wake) < 0) {
 		warn(_("read rtc alarm failed"));
@@ -340,7 +374,8 @@ static int print_alarm(struct rtcwake_control *ctl, int fd)
 	}
 	/* 0 if both UTC, or expresses diff if RTC in local time */
 	alarm += ctl->sys_time - ctl->rtc_time;
-	printf(_("alarm: on  %s"), ctime(&alarm));
+	ctime_r(&alarm, s);
+	printf(_("alarm: on  %s"), s);
 
 	return 0;
 }
@@ -402,12 +437,11 @@ int main(int argc, char **argv)
 		.clock_mode = CM_AUTO
 	};
 	char *devname = DEFAULT_RTC_DEVICE;
-	unsigned seconds = 0;
 	int suspend = SYSFS_MODE;
 	int rc = EXIT_SUCCESS;
 	int t;
 	int fd;
-	time_t alarm = 0;
+	time_t alarm = 0, seconds = 0;
 	enum {
 		OPT_DATE = CHAR_MAX + 1,
 		OPT_LIST
@@ -432,13 +466,14 @@ int main(int argc, char **argv)
 	static const ul_excl_t excl[] = {
 		{ 'a', 'l', 'u' },
 		{ 's', 't', OPT_DATE },
+		{ 0 },
 	};
 	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
 
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout);
+	close_stdout_atexit();
 
 	while ((t = getopt_long(argc, argv, "A:ahd:lm:ns:t:uVv",
 					long_options, NULL)) != EOF) {
@@ -472,16 +507,16 @@ int main(int argc, char **argv)
 			break;
 		case 's':
 			/* alarm time, seconds-to-sleep (relative) */
-			seconds = strtou32_or_err(optarg, _("invalid seconds argument"));
+			seconds = strtotime_or_err(optarg, _("invalid seconds argument"));
 			break;
 		case 't':
 			/* alarm time, time_t (absolute, seconds since epoch) */
-			alarm = strtou32_or_err(optarg, _("invalid time argument"));
+			alarm = strtotime_or_err(optarg, _("invalid time argument"));
 			break;
 		case OPT_DATE:
 		{	/* alarm time, see timestamp format from manual */
 			usec_t p;
-			if (parse_timestamp(optarg, &p) < 0)
+			if (ul_parse_timestamp(optarg, &p) < 0)
 				errx(EXIT_FAILURE, _("invalid time value \"%s\""), optarg);
 			alarm = (time_t) (p / 1000000);
 			break;
@@ -492,13 +527,13 @@ int main(int argc, char **argv)
 		case 'v':
 			ctl.verbose = 1;
 			break;
+
 		case 'V':
-			printf(UTIL_LINUX_VERSION);
-			exit(EXIT_SUCCESS);
+			print_version(EXIT_SUCCESS);
 		case 'h':
-			usage(stdout);
+			usage();
 		default:
-			usage(stderr);
+			errtryhelp(EXIT_FAILURE);
 		}
 	}
 
@@ -525,30 +560,41 @@ int main(int argc, char **argv)
 		exit(EXIT_FAILURE);
 
 	if (ctl.verbose)
-		printf(_("alarm %ld, sys_time %ld, rtc_time %ld, seconds %u\n"),
-				alarm, ctl.sys_time, ctl.rtc_time, seconds);
+		printf(_("alarm %"PRId64", sys_time %"PRId64", "
+			 "rtc_time %"PRId64", seconds %"PRIu64"\n"),
+				(int64_t) alarm, (int64_t) ctl.sys_time,
+				(int64_t) ctl.rtc_time,
+				(int64_t) seconds);
 
 	if (suspend != DISABLE_MODE && suspend != SHOW_MODE) {
 		/* perform alarm setup when the show or disable modes are not set */
 		if (alarm) {
-			if (alarm < ctl.sys_time)
-				errx(EXIT_FAILURE, _("time doesn't go backward to %s"),
-						ctime(&alarm));
-			alarm += ctl.sys_time - ctl.rtc_time;
+			if (alarm < ctl.sys_time) {
+				char s[CTIME_BUFSIZ];
+
+				ctime_r(&alarm, s);
+				errx(EXIT_FAILURE, _("time doesn't go backward to %s"), s);
+			}
+			alarm -= ctl.sys_time - ctl.rtc_time;
 		} else
 			alarm = ctl.rtc_time + seconds + 1;
 
 		if (setup_alarm(&ctl, fd, &alarm) < 0)
 			exit(EXIT_FAILURE);
 
-		if (suspend == NO_MODE || suspend == ON_MODE)
+		if (suspend == NO_MODE || suspend == ON_MODE) {
+			char s[CTIME_BUFSIZ];
+
+			ctime_r(&alarm, s);
 			printf(_("%s: wakeup using %s at %s"),
-				program_invocation_short_name, devname,
-				ctime(&alarm));
-		else
+				program_invocation_short_name, devname, s);
+		} else {
+			char s[CTIME_BUFSIZ];
+
+			ctime_r(&alarm, s);
 			printf(_("%s: wakeup from \"%s\" using %s at %s"),
-				program_invocation_short_name, ctl.mode_str, devname,
-				ctime(&alarm));
+				program_invocation_short_name, ctl.mode_str, devname, s);
+		}
 		fflush(stdout);
 		xusleep(10 * 1000);
 	}
@@ -564,18 +610,32 @@ int main(int argc, char **argv)
 		char *arg[5];
 		int i = 0;
 
-		if (ctl.verbose)
-			printf(_("suspend mode: off; executing %s\n"),
-						_PATH_SHUTDOWN);
-		arg[i++] = _PATH_SHUTDOWN;
-		arg[i++] = "-h";
-		arg[i++] = "-P";
-		arg[i++] = "now";
-		arg[i]   = NULL;
-		if (!ctl.dryrun) {
-			execv(arg[0], arg);
-			warn(_("failed to execute %s"), _PATH_SHUTDOWN);
-			rc = EXIT_FAILURE;
+		if (!access(_PATH_SHUTDOWN, X_OK)) {
+			arg[i++] = _PATH_SHUTDOWN;
+			arg[i++] = "-h";
+			arg[i++] = "-P";
+			arg[i++] = "now";
+			arg[i]   = NULL;
+		} else if (!access(_PATH_POWEROFF, X_OK)) {
+			arg[i++] = _PATH_POWEROFF;
+			arg[i]   = NULL;
+		} else {
+			arg[i] 	 = NULL;
+		}
+
+		if (arg[0]) {
+			if (ctl.verbose)
+				printf(_("suspend mode: off; executing %s\n"),
+						arg[0]);
+			if (!ctl.dryrun) {
+				execv(arg[0], arg);
+				warn(_("failed to execute %s"), arg[0]);
+				rc = EX_EXEC_ENOENT;
+			}
+		} else {
+			/* Failed to find shutdown command */
+			warn(_("failed to find shutdown command"));
+			rc = EX_EXEC_ENOENT;
 		}
 		break;
 	}
